package hu.eltesoft.modelexecution.m2m.logic.viatra

import com.google.common.base.Stopwatch
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.Behavior
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.BehaviorMatch
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.Classdef
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.ClsMatch
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.Event
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.EventMatch
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.Region
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.RegionMatch
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.Signal
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.SignalMatch
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.util.BehaviorQuerySpecification
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.util.ClsQuerySpecification
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.util.EventQuerySpecification
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.util.RegionQuerySpecification
import hu.eltesoft.modelexecution.m2m.logic.viatra.queries.util.SignalQuerySpecification
import hu.eltesoft.modelexecution.m2m.logic.viatra.util.RuleProvider
import hu.eltesoft.modelexecution.m2m.metamodel.base.ModelRoot
import java.util.HashMap
import java.util.Map
import java.util.concurrent.TimeUnit
import org.apache.log4j.Logger
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.incquery.runtime.api.GenericPatternGroup
import org.eclipse.incquery.runtime.api.IncQueryEngine
import org.eclipse.uml2.uml.Element
import org.eclipse.viatra.emf.runtime.changemonitor.ChangeDelta
import org.eclipse.viatra.emf.runtime.changemonitor.ChangeMonitor
import org.eclipse.viatra.emf.runtime.rules.batch.BatchTransformationStatements
import org.eclipse.viatra.emf.runtime.transformation.batch.BatchTransformation

import static com.google.common.base.Preconditions.*

import static extension hu.eltesoft.modelexecution.m2m.logic.viatra.util.TransformationActions.*

class ChangeListenerBatchTransformation extends MTMTransformation{
	extension Logger logger = Logger.getLogger("cps.xform.m2m.batch.viatra")

	Resource resource
	IncQueryEngine engine
	ChangeMonitor monitor
	Map<Element, ModelRoot> traceMap
	boolean firstRun = true
	
	extension RuleProvider ruleProvider
	extension BatchTransformation transformation
	extension BatchTransformationStatements statements
	
	override initialize(Resource targetResource, IncQueryEngine engine) {
		checkArgument(targetResource != null, "Target resource cannot be null!")
		checkArgument(engine != null, "Engine cannot be null!")

		this.resource = targetResource
		this.engine = engine
		
		debug("Preparing queries on engine.")
		val watch = Stopwatch.createStarted
		queryGroup.prepare(engine)
		traceMap = new HashMap

		monitor = new ChangeMonitor(engine);

		monitor.addRule(BehaviorQuerySpecification.instance)
		monitor.addRule(ClsQuerySpecification.instance)
		monitor.addRule(RegionQuerySpecification.instance)
		monitor.addRule(EventQuerySpecification.instance)
		monitor.addRule(SignalQuerySpecification.instance);
		monitor.startMonitoring

		watch.stop
		info('''Prepared queries on engine («watch.elapsed(TimeUnit.MILLISECONDS)» ms)''')
	}
	
	override execute() {
		trace('''Executing Transformation''')
		if(firstRun){
			ruleProvider = new RuleProvider(engine, resource)
			transformation = BatchTransformation::forEngine(engine)
			statements = new BatchTransformationStatements(transformation)
			resource.contents.clear;
			
			getBehaviorIncrRule(traceMap).fireAllCurrent
			getClsIncrRule(traceMap).fireAllCurrent
			getRegionIncrRule(traceMap).fireAllCurrent
			getEventIncrRule(traceMap).fireAllCurrent
			getSignalIncrRule(traceMap).fireAllCurrent
			
			transformation.dispose
			firstRun = false;
		}
		val delta = monitor.createCheckpoint
		deleteChanges(delta)
		addChanges(delta)
		updateChanges(delta)
		trace('''Execution ended: Transformation''')
	}
	private def deleteChanges(ChangeDelta delta){
		
		delta.disappeared.keySet.forEach[spec |
			delta.disappeared.get(spec).filter(BehaviorMatch).forEach[ match |
				removeElement(traceMap.get(match.behavior));
				traceMap.remove(match.behavior)
			]
		]
		
		delta.disappeared.keySet.forEach[spec |
			delta.disappeared.get(spec).filter(ClsMatch).forEach[ match |
				removeElement(traceMap.get(match.cls));
				traceMap.remove(match.cls)
			]
		]
		
		delta.disappeared.keySet.forEach[spec |
			delta.disappeared.get(spec).filter(RegionMatch).forEach[ match |
				removeElement(traceMap.get(match.region));
				traceMap.remove(match.region)
			]
		]
		
		delta.disappeared.keySet.forEach[spec |
			delta.disappeared.get(spec).filter(EventMatch).forEach[ match |
				removeElement(traceMap.get(match.event));
				traceMap.remove(match.event)
			]
		]
		
		delta.disappeared.keySet.forEach[spec |
			delta.disappeared.get(spec).filter(SignalMatch).forEach[ match |
				removeElement(traceMap.get(match.signal));
				traceMap.remove(match.signal)
			]
		]
	}
	
	private def addChanges(ChangeDelta delta) {
					
		delta.appeared.keySet.forEach[spec |
			delta.appeared.get(spec).filter(BehaviorMatch).forEach[ match |
				traceMap.put(match.behavior, match.addElements(engine,resource))
			]
		]
		
		delta.appeared.keySet.forEach[spec |
			delta.appeared.get(spec).filter(ClsMatch).forEach[ match |
				traceMap.put(match.cls, match.addElements(engine,resource))
			]
		]
		
		delta.appeared.keySet.forEach[spec |
			delta.appeared.get(spec).filter(RegionMatch).forEach[ match |
				traceMap.put(match.region, match.addElements(engine,resource))
			]
		]
		
		delta.appeared.keySet.forEach[spec |
			delta.appeared.get(spec).filter(EventMatch).forEach[ match |
				traceMap.put(match.event, match.addElements(engine,resource))
			]
		]
		
		delta.appeared.keySet.forEach[spec |
			delta.appeared.get(spec).filter(SignalMatch).forEach[ match |
				traceMap.put(match.signal, match.addElements(engine,resource))
			]
		]
	}
	
	private def updateChanges(ChangeDelta delta) {
				
		delta.updated.keySet.forEach[spec |
			delta.updated.get(spec).filter(BehaviorMatch).forEach[ match |
				removeElement(traceMap.get(match.behavior));
				traceMap.remove(match.behavior)
				traceMap.put(match.behavior, match.addElements(engine,resource))
			]
		]
		
		delta.updated.keySet.forEach[spec |
			delta.updated.get(spec).filter(ClsMatch).forEach[ match |
				removeElement(traceMap.get(match.cls));
				traceMap.remove(match.cls)
				traceMap.put(match.cls, match.addElements(engine,resource))
			]
		]
		
		delta.updated.keySet.forEach[spec |
			delta.updated.get(spec).filter(RegionMatch).forEach[ match |
				removeElement(traceMap.get(match.region));
				traceMap.remove(match.region)
				traceMap.put(match.region, match.addElements(engine,resource))
			]
		]
		
		delta.updated.keySet.forEach[spec |
			delta.updated.get(spec).filter(EventMatch).forEach[ match |
				removeElement(traceMap.get(match.event));
				traceMap.remove(match.event)
				traceMap.put(match.event, match.addElements(engine,resource))
			]
		]
		
		delta.updated.keySet.forEach[spec |
			delta.updated.get(spec).filter(SignalMatch).forEach[ match |
				removeElement(traceMap.get(match.signal));
				traceMap.remove(match.signal)
				traceMap.put(match.signal, match.addElements(engine,resource))
			]
		]
	}
	
	override dispose() {
		monitor.dispose
	}
	
	private def getQueryGroup() {
		GenericPatternGroup.of(
			// Insert patterns here
			Behavior.instance(),
			Classdef.instance(),
			Event.instance(),
			Region.instance(),
			Signal.instance()
		);
	}
	
}